"""
TABI - UCSC iGEM 2023

StealthGenome: Prep Genome for Stealth and Codon Frequency analysis

PlasmidParser: Identifies Stealth-mutable regions of an annotated plasmid

Author: Tyler Gaw
"""

from ._SeqParser import *